#include "LineSrcNumItgSimulator.h"
#include "PasquillGifford.h"
#include "simpson.h"
#include "../share/common.h"
#include "../share/WildMagic.h"

void LineSrcNumItgSimulator::scan()
{
	ML_TRACE("LineSrcNumItgSimulator::scan");

	isTimeSeries() ? attachMetData(new MetTimeSeries()) : attachMetData(new Met());
	attachRcpData(new Rcp());
	attachSrcData(new LineSrc());
	attachSgmData(new PasquillGifford());
	attachPatData(new Pat());
	attachRanData(new Ran());
	attachBlgData(new Blg());
	Simulator::scan();

	ptr_sgm->scan();

	num_itg = new Simpson(this);
	num_itg->scan();
}

void LineSrcNumItgSimulator::print()
{
	ML_TRACE("LineSrcNumItgSimulator::print");

	Simulator::print();
	ptr_sgm->  print();
	num_itg->  print();
}

void LineSrcNumItgSimulator::lists()
{
	ML_TRACE("LineSrcNumItgSimulator::lists");

	Simulator::lists(); const ListLevel level = getListLevel();
	if(level == FULL) ptr_sgm->lists();
	if(level == FULL) num_itg->lists();
}

double LineSrcNumItgSimulator::integrand(double x, bool dirac)
{
	ML_TRACE("LineSrcNumItgSimulator::integrand");

	// Placeholders for coordinates
	double _x_, _y_, _z_, _h_; ptr_rcp->getCoor(_x_, _y_, _z_, _h_);

	const double sz = ptr_sgm->z(x + getVirtual4Vert());
	const double ef = (sz > 0.0 ? exp(-0.5*pow((_z_ + dh())/sz, 2)) + exp(-0.5*pow((_z_ - dh())/sz, 2)) : 0.0);
	const double dn = sqrt(2*M_PI)*M_PI/8*x*sz*getModelled_ws();

	return (dn > 0 ? (!dirac ? getEmi()*ef/dn/getAngleCosine() : getEmi()*ef/dn*getLength()) : 0.0);
}


double LineSrcNumItgSimulator::equation(PlumePuff pp)
{
	ML_TRACE("LineSrcNumItgSimulator::equation");

	if(pp == PLUME) return 0;

	// Placeholders for coordinates
	double _x_, _y_, _z_, _h_; ptr_rcp->getCoor(_x_, _y_, _z_, _h_);

	// Placeholders for Turner's parameters
	double _alp_, _gam_ ; ptr_sgm->getTurners(_alp_, _gam_);

	const double y_1 = getInfPuff(); const double y_2 = getSupPuff(); const double x_x = getDstPuff();

	const double h_eff = dh();
	const double eta1  = sqrt(x_x*x_x + _alp_*_alp_/_gam_/_gam_*(h_eff - _z_)*(h_eff - _z_));
	const double eta2  = sqrt(x_x*x_x + _alp_*_alp_/_gam_/_gam_*(h_eff + _z_)*(h_eff + _z_));

	const double fact1 = getEmi()*pow(2.0*M_PI, -1.5)/_gam_;
	const double fact2 = (atan(y_2/eta1) - atan(y_1/eta1))/eta1;
	const double fact3 = (atan(y_2/eta2) - atan(y_1/eta2))/eta2;

	return fact1*(fact2 + fact3);
}

void LineSrcNumItgSimulator::exec(bool new_met, bool new_src)
{
	ML_TRACE("LineSrcNumItgSimulator::exec");

	double res;
	switch(assembleSimulationProperties(new_met, new_src))
	{
	case 0:
		res = 0;							// Out of range
		break;
	case 1:
		res = num_itg->run();				// Execute ordinary numerical integration
		break;
	case 2:
		res = integrand(getInf(), true);	// Dirac's delta function
		break;
	case 3:
		res = equation(PUFF);				// Under the calm condition
		break;
	}
	setResult(Rcp::CON, res); setResult(Rcp::DEP, 0); setResult(Rcp::FLG, (ptr_sgm->validated() ? 1 : 0));
}

int LineSrcNumItgSimulator::assembleSimulationProperties(bool new_met, bool new_src)
{
	ML_TRACE("LineSrcNumItgSimulator::assembleSimulationProperties");

	if(new_met)
	{
		double _wd_, _ws_, _tp_; int _st_;
		ptr_met->getMet(_wd_, _ws_, _st_, _tp_);

		if(_wd_ < 0.0 || _ws_ <= 0.4 || fabs(_ws_ - 0.4) < 1.0e-5)
		{
			setCalmFlag(true);

		}
		else
		{
			setCalmFlag(false);
		}

		setOriginal_wd(_wd_);
		setOriginal_ws(_ws_);
		setModelled_st(reclassST(_st_));
		ptr_sgm->setStability(getModelled_st());
	}
	if(new_src || new_met)
	{
		// Placeholders for line source
		Link    _lnk_;
		double  _emi_;
		const double percent = ptr_src->getLineSrc(_lnk_, _emi_);
		setEmi(_emi_*percent*0.01);
		setLnk(_lnk_);


		// Placeholders for link
		int      _ide_;
		Vector3d _beg_;
		Vector3d _end_;
		double   _wid_;
		string   _des_;
		_lnk_.getLink(_ide_, _beg_, _end_, _wid_, _des_);
		setIde(_ide_);
		setBeg(_beg_);
		setEnd(_end_);
		setWid(_wid_);
		setDes(_des_);

		if((getListLevel() == Simulator::MODERATE || getListLevel() == Simulator::FULL))
		{
			CLST << "[" << setw(3) << _ide_ << ":" << setw(3) << floor(percent + 0.5) << "]";
		}


		// Evaluate initial diffusion effects
		double sol;
		const double sy0 = getSy0(); ptr_sgm->sySolve(sol, sy0); setVirtual4Latr(sol);
		const double sz0 = getSz0(); ptr_sgm->szSolve(sol, sz0); setVirtual4Vert(sol);

		//
		// Modify wind speed dependent on each source height
		//
		// If the modified wind speed is less than 1.0, code resets the wind speed to 1.0 m/s
		// because of plume formula usage limitation.
		//
		const double ws_modified_with_power_law = getOriginal_ws() * wsMltfactor();
		setModelled_ws(max(1.0, ws_modified_with_power_law));
	}

	if(getEmi() == 0.0) return 0;

	// Placeholder for local coordinates
	double x_loc;
	double y_loc;
	setModelled_wd(getOriginal_wd() + wdAddfactor());	// For randomized wind direction.
 	getLocalCoor(x_loc, y_loc, getBeg().X(), getBeg().Y()); const Vector2d& beg_loc = Vector2d(x_loc, y_loc);
	getLocalCoor(x_loc, y_loc, getEnd().X(), getEnd().Y()); const Vector2d& end_loc = Vector2d(x_loc, y_loc);

	if(isCalm())
	{
		const Vector2d e_e = (end_loc - beg_loc)/(end_loc - beg_loc).Length();
		const Vector2d E_E = Vector2d(-e_e.Y(), e_e.X());
		double y_1 = beg_loc.Dot(e_e); double y_2 = end_loc.Dot(e_e); if(y_1 > y_2){const double bak = y_2; y_2 = y_1; y_1 = bak;}
		double x_x = beg_loc.Dot(E_E);
		setInfPuff(y_1); setSupPuff(y_2); setDstPuff(x_x);
		return 3;
	}

	// Placeholders for line objects
	Segment2d seg; Ray2d ray1,ray2;
	seg.Origin()    = beg_loc;
	seg.Direction() = end_loc - beg_loc;

	// Placeholder for FindIntersection
	int  quant1, quant2; double t1[2], t2[2];
	ray1.Origin() = Vector2d::ZERO; ray1.Direction() = Vector2d(-getTrigHalfOpenAngle(COS),  getTrigHalfOpenAngle(SIN));
	const bool test1 = FindIntersection(ray1, seg, quant1, t1);
	ray2.Origin() = Vector2d::ZERO; ray2.Direction() = Vector2d(-getTrigHalfOpenAngle(COS), -getTrigHalfOpenAngle(SIN));
	const bool test2 = FindIntersection(ray2, seg, quant2, t2);


	Vector2d _beg, _end;
	if(test1 || test2)
	{
		if(quant1 == 2 || quant2 == 2)
		{
			if(quant1 == 2)
			{
				_beg = t1[0]*ray1.Direction();
				_end = t1[1]*ray1.Direction();
			}
			else
			{
				_beg = t2[0]*ray2.Direction();
				_end = t2[1]*ray2.Direction();
			}
		}
		else
		{
			if(test1 && test2)
			{
				_beg = t1[0]*ray1.Direction();
				_end = t2[0]*ray2.Direction();
			}
			else if(test1)
			{
				_beg = t1[0]*ray1.Direction();
				_end = (beg_loc.X() <= -getTrigHalfOpenAngle(COS)*beg_loc.Length() ? beg_loc : end_loc);
			}
			else
			{
				_beg = t2[0]*ray2.Direction();
				_end = (beg_loc.X() <= -getTrigHalfOpenAngle(COS)*beg_loc.Length() ? beg_loc : end_loc);
			}
		}
	}
	else
	{
		if(beg_loc.X() <= -getTrigHalfOpenAngle(COS)*beg_loc.Length())
		{
			_beg = beg_loc;
			_end = end_loc;
		}
		else
		{
			return 0;
		}
	}

	const double beg_coord_x = -_beg.X();
	const double end_coord_x = -_end.X();
	const double beg_coord_y =  _beg.Y();
	const double end_coord_y =  _end.Y();
	const double len       = (_end - _beg).Length();
	const double cosine    = fabs((end_coord_x - beg_coord_x)/len);

	setAngleCosine(cosine);
	if(beg_coord_x < end_coord_x)
	{
		setInf  (beg_coord_x);
		setSup  (end_coord_x);
		setInf_y(beg_coord_y);
		setSup_y(end_coord_y);
	}
	else
	{
		setInf  (end_coord_x);
		setSup  (beg_coord_x);
		setInf_y(end_coord_y);
		setSup_y(beg_coord_y);
	}
	setLength(len);

	return (fabs(getInf() - getSup()) < 1.0e-3 ? 2 : 1);
}

double LineSrcNumItgSimulator::wsMltfactor () const
{
	ML_TRACE("LineSrcNumItgSimulator::wsMltfactor");

	//
	// Rather experimental implementation.
	// Reference wind height fixed to 10 meters.
	//
	return (ptr_met->getAnemometerHeight() <= 1.0e-6 ? 1 : pow(10.0/ptr_met->getAnemometerHeight(), power4ws(getModelled_st(), RURAL)));
}

