#include "PointDropletSrcSimulator.h"
#include "PasquillGifford.h"

void PointDropletSrcSimulator::scan()
{
	ML_TRACE("PointDropletSrcSimulator::scan");

	droplet->scan();

	isTimeSeries() ? attachMetData(new MetTimeSeries()) : attachMetData(new Met());
	attachRcpData(new Rcp());
	attachSrcData(new PointDropletSrc(droplet));
	attachSgmData(new PasquillGifford());
	attachPatData(new Pat());
	attachRanData(new Ran());
	attachBlgData(new Blg());

	ptr_blg->scan(); Simulator::scan(); if(ptr_rcp->getTerrainOpt()) attachAdjData(new TerrainAdjustment());
}

double PointDropletSrcSimulator::getGravitationalSettlingVelocity() const
{
	ML_TRACE("PointDropletSrcSimulator::getGravitationalSettlingVelocity");

	const double rho_p = getDropletDensity();
	const double rho_a = 1.25;
	const double mu    = 1.5e-5;
	const double r     = getDropletRadius();
	return (getDropletDrag() < 1.0e-6 ? 0 : (2*r*r*rho_p*9.81/9/mu/rho_a)/getDropletDrag());
}

double PointDropletSrcSimulator::getDepositionVelocity() const
{
	ML_TRACE("PointDropletSrcSimulator::getDepositionVelocity");

	/*
	return (getDropletRadius() == 0.0 ? 0.0 : getGravitationalSettlingVelocity() + 0.006*getModelled_ws());
	*/


	// Revised implementation. (2006/06/14)
	const double ws10 = getOriginal_ws()*wsMltfactorZ(10.0);
	// CLOG << "(orig_ws, wltfact, ws10) = " << getOriginal_ws() << ',' << wsMltfactorZ(10.0) << ',' << ws10 << endl;
	return (getDropletRadius() == 0.0 ? 0.0 : getGravitationalSettlingVelocity() + 0.006*ws10);
}

double PointDropletSrcSimulator::dh() const
{
	ML_TRACE("PointDropletSrcSimulator::dh");

	if(getPollutantProperty() == GAS)
		return getDhAtStackPosition();	// Rewritten for code readability.
	else
		return max(-getCtr().Z(), getDhAtStackPosition() - getGravitationalSettlingVelocity()*getDownwindDistance()/getModelled_ws());
}

double PointDropletSrcSimulator::equation(PlumePuff pp)
{
	ML_TRACE("PointDropletSrcSimulator::equation");

	// Placeholders for receptor coordinates
	double _x_,_y_,_z_,_h_;	ptr_rcp->getCoor(_x_, _y_, _z_, _h_);

	const double d_h = dh();
	const double h_s = getCtr().Z(); double h_eff = h_s + d_h;

	if(ptr_rcp->getTerrainOpt())
	{
		ptr_adj->setDebugInfos(getIde(), _x_, _y_);
		ptr_adj->setParameters(d_h, h_s, getAlt(), _h_);
		h_eff  = ptr_adj->simpl();
	}

	if(pp == PLUME)
	{
		const double xx = getDownwindDistance();
		const double xv = getVirtual4Latr();
		const double sy = ptr_sgm->y(xx + xv);
		const double sz = ptr_sgm->z(xx + xv);
		const double fact1 = 0.5*getEmi()/M_PI/sy/sz/getModelled_ws();
		const double fact2 = exp(- 0.5*pow(getCrosswindDistance()/sy, 2));
		const double atten = (getPollutantProperty() == GAS ? 1.0 : alpha(h_eff));
		const double fact3 = exp(- 0.5*pow((_z_ - h_eff)/sz, 2)) + atten*exp(- 0.5*pow((_z_ + h_eff)/sz, 2));
		const double fact4 = exp(- 0.5*pow((    - h_eff)/sz, 2)) + atten*exp(- 0.5*pow((      h_eff)/sz, 2));

		if(VALIDATION_Q)	// Normally VALIDATION_Q is set to false.
		{
			CLST << "\n\n";
			CLST << setw(30) << "xx,sy,sy',sz=";
			CLST << setw(10) << xx;
			CLST << setw(10) << sy << setw(10) << sy/ptr_sgm->getDilutionFactor();
			CLST << setw(10) << sz << endl;
			CLST << setw(30) << "emi,ws=";
			CLST << setw(10) << getEmi();
			CLST << setw(10) << getModelled_ws() << endl;
			CLST << "\n\n";
		}

		if(TRACE_Q)
		{
			CLOG << "fact1=" << fact1 << ' ';
			CLOG << "fact2=" << fact2 << ' ';
			CLOG << "fact3=" << fact3 << endl;

			CLOG << "he=" << h_eff << ' ';
			CLOG << "xx=" << xx << ' ';
			CLOG << "xv=" << xv << ' ';
			CLOG << "sy=" << sy << ' ';
			CLOG << "sz=" << sz << ' ';
			CLOG << "em=" << getEmi() << ' ';
			CLOG << "ws=" << getModelled_ws() << endl;
		}

		setReceptorConcentrationAtZ(fact1*fact2*fact3);
		setReceptorConcentrationAt0(fact1*fact2*fact4);

		return fact1*fact2*fact3;	// Return the receptor level concentration only
	}
	else
	{
		return PointSrcSimulator::equation(PUFF);
	}
}

double PointDropletSrcSimulator::alpha(double h_e)
{
	ML_TRACE("PointDropletSrcSimulator::alpha");

	const double dx = 1.0;
	const double vs = getGravitationalSettlingVelocity();
	const double vd = getDepositionVelocity();
	const double xx = getDownwindDistance();
	const double uu = getModelled_ws();
	const double sz = ptr_sgm->z(xx + getVirtual4Vert());
	const double ds = ptr_sgm->z(xx + getVirtual4Vert() + dx) - sz;
	const double dsdx = ds/dx;
	return max(1.0 - 2*vd/(vs + vd + (uu*h_e - vs*xx)/sz*dsdx), 0.0);
}

void PointDropletSrcSimulator::exec(bool new_met, bool new_src)
{
	ML_TRACE("PointDropletSrcSimulator::exec");

	con = dep = 0;
	switch(assembleSimulationProperties(new_met, new_src))
	{
	case 0:
		setResult(Rcp::CON, 0);	// Out of range
		setResult(Rcp::DEP, 0);
		break;
	case 1:
		{
			droplet->head();
			for(int i = 1; i <= droplet->getNoElms(); i++)
			{
				const double bd = ptr_src->getBulkDensity(i);
				const double wr = ptr_src->getWeightRatio(i);
				if(fabs(wr) > 1.0e-6)
				{
					double medium_drg; ptr_src->getMediumProperties(medium_drg);
					setDropletDensity   (i == 1 ? 0   : 1000*bd);					// Checking whether gaseous or particulate
					setDropletDrag      (medium_drg);
					setDropletDiameter  (i == 1 ? 0   : droplet->getDiameter());	// Checking whether gaseous or particulate
					setPollutantProperty(i == 1 ? GAS : DROPLET);

					evalAndAddThisDropletContrib(wr);
				}
				droplet->advance();
			}
			droplet->head();
			setResult(Rcp::CON, con);
			setResult(Rcp::DEP, dep);
			break;
		}
	case 2:
		setPollutantProperty(GAS);
		setResult(Rcp::CON, equation(PUFF));
		setResult(Rcp::DEP, 0);
		break;
	}
	setResult(Rcp::FLG, (ptr_sgm->validated() ? 1 : 0));
}

void PointDropletSrcSimulator::evalAndAddThisDropletContrib(double wgt)
{
	ML_TRACE("PointDropletSrcSimulator::evalAndAddThisDropletContrib");

	const double placeholder = equation(PLUME);
	con += wgt*getReceptorConcentrationAtZ();
	if(getPollutantProperty() == DROPLET) dep += wgt*getReceptorConcentrationAt0()*getDepositionVelocity();
}

void PointDropletSrcSimulator::print()
{
	ML_TRACE("PointDropletSrcSimulator::print");

	PointSrcSimulator::print(); ptr_blg->print();
}

void PointDropletSrcSimulator::lists()
{
	ML_TRACE("PointDropletSrcSimulator::lists");

	PointSrcSimulator::lists(); const ListLevel level = getListLevel();
	if(level == MODERATE || level == FULL) ptr_blg->lists();
}
