#include "PointSrcSimulator.h"
#include "PasquillGifford.h"


void PointSrcSimulator::scan()
{
	ML_TRACE("PointSrcSimulator::scan");

	isTimeSeries() ? attachMetData(new MetTimeSeries()) : attachMetData(new Met());
	attachRcpData(new Rcp());
	attachSrcData(new PointSrc());
	attachSgmData(new PasquillGifford());
	attachPatData(new Pat());
	attachRanData(new Ran());
	attachBlgData(new Blg());

	Simulator::scan(); if(ptr_rcp->getTerrainOpt()) attachAdjData(new TerrainAdjustment());
}

void PointSrcSimulator::print()
{
	ML_TRACE("PointSrcSimulator::print");

	Simulator::print();
}

void PointSrcSimulator::lists()
{
	ML_TRACE("PointSrcSimulator::lists");

	Simulator::lists();
}

double PointSrcSimulator::equation(PlumePuff pp)
{
	// Placeholders for receptor coordinates
	double _x_,_y_,_z_,_h_;	ptr_rcp->getCoor(_x_, _y_, _z_, _h_);

	const double d_h = dh(); const double h_s = getCtr().Z(); double h_eff = h_s + d_h;
	if(ptr_rcp->getTerrainOpt())
	{
		ptr_adj->setDebugInfos(getIde(), _x_, _y_);
		ptr_adj->setParameters(d_h, h_s, getAlt(), _h_);
		h_eff = ptr_adj->simpl();
	}

	if(pp == PLUME)
	{
		ML_TRACE("PointSrcSimulator::equation PLUME");

		const double sy = ptr_sgm->y(getDownwindDistance() + getVirtual4Latr());
		const double sz = ptr_sgm->z(getDownwindDistance() + getVirtual4Vert());
		const double fact1 = 0.5*getEmi()/M_PI/sy/sz/getModelled_ws();
		const double fact2 = exp(- 0.5*pow(getCrosswindDistance()/sy, 2));
		const double fact3 = exp(- 0.5*pow((_z_ - h_eff)/sz, 2)) + exp(- 0.5*pow((_z_ + h_eff)/sz, 2));
		return fact1*fact2*fact3;
	}
	else
	{
		ML_TRACE("PointSrcSimulator::equation PUFF");

		const double xx = getDownwindDistance();
		const double yy = getCrosswindDistance();
		const double rr = sqrt(xx*xx + yy*yy);

		double _alp_, _gam_; ptr_sgm->getTurners(_alp_, _gam_);

		const double fact1 = getEmi()*pow(2.0*M_PI, -1.5)/_gam_;
		const double fact2 = rr*rr + _alp_*_alp_/_gam_/_gam_*(h_eff - _z_)*(h_eff - _z_);
		const double fact3 = rr*rr + _alp_*_alp_/_gam_/_gam_*(h_eff + _z_)*(h_eff + _z_);

		if(TRACE_Q)
		{
			CLOG << "_alp_=" << _alp_    << ' ';
			CLOG << "_gam_=" << _gam_    << ' ';
			CLOG << "   rr=" << rr       << ' ';
			CLOG << "   em=" << getEmi() << endl;
			CLOG << "h_eff=" << h_eff    << endl;
			CLOG << "h_s="   << h_s      << endl;
			CLOG << "d_h="   << d_h      << endl;
			CLOG << "_z_="   << _z_      << endl;
			CLOG << "fact1=" << fact1    << endl;
			CLOG << "fact2=" << fact2    << endl;
			CLOG << "fact3=" << fact3    << endl;
			CLOG << "res="   << fact1*(1.0/fact2 + 1.0/fact3) << endl;
		}

		return fact1*(1.0/fact2 + 1.0/fact3);
	}
}

void PointSrcSimulator::exec(bool new_met, bool new_src)
{
	ML_TRACE("PointSrcSimulator::exec");

	switch(assembleSimulationProperties(new_met, new_src))
	{
	case 0:
		setResult(Rcp::CON, 0);		// Out of range
		break;
	case 1:
		setResult(Rcp::CON, equation(PLUME));
		break;
	case 2:
		setResult(Rcp::CON, equation(PUFF));
		break;
	}
	setResult(Rcp::DEP, 0); setResult(Rcp::FLG, (ptr_sgm->validated() ? 1 : 0));
}

int PointSrcSimulator::assembleSimulationProperties(bool new_met, bool new_src)
{
	ML_TRACE("PointSrcSimulator::assembleSimulationProperties");

	if(new_met)
	{
		double _wd_, _ws_, _tp_; int _st_;
		ptr_met->getMet(_wd_, _ws_, _st_, _tp_);

		if(_wd_ < 0.0 || _ws_ <= 0.4 || fabs(_ws_ - 0.4) < 1.0e-5)
		{
			setCalmFlag(true);

		}
		else
		{
			setCalmFlag(false);
		}

		setOriginal_wd(_wd_);
		setOriginal_ws(_ws_);
		setModelled_st(reclassST(_st_));
		ptr_sgm->setStability(getModelled_st());
	}
	if(new_src || new_met)
	{

		// Placeholders for source object
		Ring _rin_; bool _stk_, _plm_; double _emi_,_vel_,_gas_,_tpr_;
		const double percent = ptr_src->getPointSrc(_rin_, _stk_, _plm_, _emi_, _vel_, _gas_, _tpr_);
		setStk(_stk_);
		setPlm(_plm_);
		setEmi(_emi_*percent*0.01);
		setVel(_vel_*percent*0.01);
		setGas(_gas_*percent*0.01);
		setTpr(_tpr_);

		// Placeholders for ring object
		int  _ide_; Vector3d _ctr_; double _dia_,_hhh_; string _des_;
		_rin_.getRing(_ide_, _ctr_, _dia_, _hhh_, _des_);
		setIde(_ide_);
		setCtr(_ctr_);
		setDia(_dia_);
		setAlt(_hhh_);

		if((getListLevel() == Simulator::MODERATE || getListLevel() == Simulator::FULL))
		{
			CLST << "[" << setw(3) << _ide_ << ":" << setw(3) << floor(percent + 0.5) << "]";
		}

		//
		// If the modified wind speed is less than 1.0, code resets the wind speed to 1.0 m/s
		// because of plume formula usage limitation.
		//
		double ws_modified_with_power_law = getOriginal_ws()*wsMltfactor();
		ws_modified_with_power_law = max(1.0, ws_modified_with_power_law);
		evaluateDhAtStackPosition(ws_modified_with_power_law);
	}

	if(getEmi() == 0.0) return 0;

	// Placeholder for local coordinates
	double x_loc; double y_loc;
	setModelled_wd(getOriginal_wd() + wdAddfactor());	// For randomized wind direction.
	getLocalCoor(x_loc, y_loc, getCtr().X(), getCtr().Y());

	setDownwindDistance(fabs(x_loc)); setCrosswindDistance(fabs(y_loc));

	if(isCalm())
	{
		return 2;
	}
	else if(getCrosswindDistance() > getDownwindDistance()*getTrigHalfOpenAngle(TAN) || x_loc > 0)
	{
		return 0;
	}
	else
	{
		return 1;
	}
}

double PointSrcSimulator::wsMltfactor () const
{
	ML_TRACE("PointSrcSimulator::wsMltfactor");

	// Placeholders for source object
	Ring _rin_; bool _stk_, _plm_; double _emi_,_vel_,_gas_,_tpr_;
	ptr_src->getPointSrc(_rin_, _stk_, _plm_, _emi_, _vel_, _gas_, _tpr_);

	// Placeholders for ring object
	int  _ide_; Vector3d _ctr_; double _dia_,_hhh_; string _des_;
	_rin_.getRing(_ide_, _ctr_, _dia_, _hhh_, _des_);

	return (ptr_met->getAnemometerHeight() <= 1.0e-6 ? 1 : pow(_ctr_.Z()/ptr_met->getAnemometerHeight(), power4ws(getModelled_st(), RURAL)));
}

double PointSrcSimulator::stacktip(double ws_modified_with_power_law) const
{
	ML_TRACE("PointSrcSimulator::stacktip");

	if(!getStk() || getEmi() == 0.0) return 0;

	double v_s;

	if(false){/* dummy */}
	else if(getDia() < 1.0e-6)
	{
		return 0;
	}
	else if(getVel() > 0)
	{
		v_s = getVel();
	}
	else
	{
		const double area = 0.25*M_PI*getDia()*getDia();
		const double gvol = getTpr()/273.15*getGas();
		v_s = gvol/area;
	}
	return max(min(2.0*getDia()*(v_s/ws_modified_with_power_law - 1.5), 0), - getCtr().Z());
}

double PointSrcSimulator::concawe(double ws_modified_with_power_law) const
{
	ML_TRACE("PointSrcSimulator::concawe");

	if(!getPlm() || getEmi() == 0.0) return 0;

	double q_q;

	if(false){/* dummy */}
	else if(getTpr() < 1.0e-6)
	{
		return 0;
	}
	else if(getGas() > 0)
	{
		q_q = getGas();
	}
	else
	{
		const double area = 0.25*M_PI*getDia()*getDia();
		q_q = 273.15/getTpr()*area*getVel();
	}

	const double rho = 1293;
	const double c_p = 0.24;
	const double q_h = max(rho*c_p*q_q*(getTpr() - 288.15), 0.0);

	return 0.175*sqrt(q_h)*pow(ws_modified_with_power_law, -0.75);
}

double PointSrcSimulator::briggs() const
{
	ML_TRACE("PointSrcSimulator::briggs");

	if(!getPlm()) return 0;

	double q_q;

	if(false){/* dummy */}
	else if(getTpr() < 1.0e-6)
	{
		return 0;
	}
	else if(getGas() > 0)
	{
		q_q = getGas();
	}
	else
	{
		const double area = 0.25*M_PI*getDia()*getDia();
		q_q = 273.15/getTpr()*area*getVel();
	}

	const double rho = 1293;
	const double c_p = 0.24;
	const double q_h = max(rho*c_p*q_q*(getTpr() - 288.15), 0.0);

	return 1.4*pow(q_h, 0.25)*pow(getDerivativePotentialTemperature(), -0.375);
}

double PointSrcSimulator::dh() const
{
	ML_TRACE("PointSrcSimulator::dh");

	return getDhAtStackPosition();
}

void PointSrcSimulator::evaluateDhAtStackPosition(double ws_modified_with_power_law)
{
	ML_TRACE("PointSrcSimulator::evaluateDhAtStackPosition");



	double p_rise;


	if(isCalm())
	{
		p_rise = briggs();
	}
	else
	{
		const double s = stacktip(ws_modified_with_power_law);
		const double c =  concawe(ws_modified_with_power_law);
		p_rise = s + c;
	}

	setModelled_ws(isCalm() ? 0.0 : max(1.0, ws_modified_with_power_law));

	const ListLevel level = getListLevel();

	if(level == MODERATE || level == FULL)
	{
		CLST << "<" << setw(5) << 0.1*floor(10*p_rise + 0.5) << ">";
	}


	setDhAtStackPosition(p_rise);
}
