#ifndef SIMULATOR_H
#define SIMULATOR_H

#include "MetTimeSeries.h"
#include "src.h"
#include "rcp.h"
#include "sgm.h"
#include "pat.h"
#include "ran.h"
#include "blg.h"
#include "TerrainAdjustment.h"
#include "../share/TimeCode.h"

class SimulationFactory;

class Simulator
{
	double* lt_res; double* lt_dep; double* lt_flg; int* lt_nrs; bool is_time_series; TimeCode* tim;
	double* st_res; double* st_dep; double* st_flg;

	double original_wd;
	double original_ws;
	int    original_st;

	double modelled_wd;
	double modelled_ws;
	int    modelled_st;

	double virtual_latr;
	double virtual_vert;

	double sz0;
	double sy0;

	double open_angle;

	double concentration;
	double deposition;
	double validation;

	bool   is_day;
	bool   is_calm;

	bool   concentration_unit_flag;
	string casename_prefix;
	int    progress_peek_step;
	int    minimum_iterations;
	int    maximum_iterations;
	double tolerance_ratio;
protected:
	Met*  ptr_met;
	Src*  ptr_src;
	Rcp*  ptr_rcp;
	Sgm*  ptr_sgm;
	Blg*  ptr_blg;
	Pat*  ptr_pat;
	Ran*  ptr_ran;
	TerrainAdjustment* ptr_adj;

	bool  scan_met,  scan_rcp,  scan_src;
	bool print_met, print_rcp, print_src;

protected:
	enum Sgn  {PLUS, MINUS};
	enum Trig {COS, SIN, TAN};

	virtual void attachMetData(Met* ptr_met_){ptr_met = ptr_met_;}
	virtual void attachRcpData(Rcp* ptr_rcp_){ptr_rcp = ptr_rcp_;}
	virtual void attachSrcData(Src* ptr_src_){ptr_src = ptr_src_;}
	virtual void attachSgmData(Sgm* ptr_sgm_){ptr_sgm = ptr_sgm_;}
	virtual void attachPatData(Pat* ptr_pat_){ptr_pat = ptr_pat_;}
	virtual void attachRanData(Ran* ptr_ran_){ptr_ran = ptr_ran_;}
	virtual void attachBlgData(Blg* ptr_blg_){ptr_blg = ptr_blg_;}
	virtual void attachAdjData(TerrainAdjustment* ptr_adj_){ptr_adj = ptr_adj_;}

	virtual void detachMetData(){if(ptr_met != NULL){delete ptr_met; ptr_met = NULL;}}
	virtual void detachRcpData(){if(ptr_rcp != NULL){delete ptr_rcp; ptr_rcp = NULL;}}
	virtual void detachSrcData(){if(ptr_src != NULL){delete ptr_src; ptr_src = NULL;}}
	virtual void detachSgmData(){if(ptr_sgm != NULL){delete ptr_sgm; ptr_sgm = NULL;}}
	virtual void detachPatData(){if(ptr_pat != NULL){delete ptr_pat; ptr_pat = NULL;}}
	virtual void detachRanData(){if(ptr_ran != NULL){delete ptr_ran; ptr_ran = NULL;}}
	virtual void detachBlgData(){if(ptr_blg != NULL){delete ptr_blg; ptr_blg = NULL;}}
	virtual void detachAdjData(){if(ptr_adj != NULL){delete ptr_adj; ptr_adj = NULL;}}

	virtual void getLocalCoor(double& x_loc, double& y_loc, double x_glo, double y_glo) const;

	// Analytic form of diffusion equation
	enum PlumePuff {PLUME, PUFF};
	virtual double equation(PlumePuff pp = PLUME){return 0;}

	// Integrand for numerical integration
	virtual double integrand(double x, bool dirac = false){return 0;}
	virtual void   setInf(double /* inf_ */){}
	virtual double getInf() const {return 0;}
	virtual void   setSup(double /* sup_ */){}
	virtual double getSup() const {return 0;}

	virtual int  assembleSimulationProperties(bool new_met, bool new_src){return 0;}

	virtual void   setCalmFlag(bool is_calm_) {is_calm = is_calm_;}
	virtual bool   isCalm() const {return is_calm;}

	// Modifiers dependent on each simulation case
	virtual double dh()               const {return 0;}		// N.B. 1. Minus sign acceptable. 2. This function gives the net plume rise height (Not real stack height!).
	virtual double wsMltfactor ()     const {return 1;}
	virtual double wsMltfactorZ(double z_) const
	{
		ML_TRACE("PointSrcSimulator::wsMltfactor(z)");
		return (ptr_met->getAnemometerHeight() <= 1.0e-6 ? 1 : pow(z_/ptr_met->getAnemometerHeight(), power4ws(getModelled_st(), RURAL)));
	}	// This function on 2006/06/14 added.

	virtual double wdAddfactor ()     const {double ra; ptr_ran->getRan(ra); return ra;}
	virtual int    reclassST(int st_);
	virtual void   setDayFlag(bool is_day_){is_day = is_day_;}
	virtual bool   isDay() const {return is_day;}
	virtual double getDerivativePotentialTemperature() const {return (isDay() ? 0.003 : 0.010);}
	virtual void   setVirtual4Latr(double virtual_latr_){virtual_latr = virtual_latr_;}		// Virtual point distance for lateral  dimension
	virtual double getVirtual4Latr()  const {return virtual_latr;}							// Virtual point distance for lateral  dimension
	virtual void   setVirtual4Vert(double virtual_vert_){virtual_vert = virtual_vert_;}		// Virtual point distance for vertical dimension
	virtual double getVirtual4Vert()  const {return virtual_vert;}							// Virtual point distance for vertical dimension

	virtual void   setOriginal_wd(double original_wd_){original_wd = original_wd_;}
	virtual double getOriginal_wd() const {return original_wd;}
	virtual void   setOriginal_ws(double original_ws_){original_ws = original_ws_;}
	virtual double getOriginal_ws() const {return original_ws;}
	virtual void   setOriginal_st(int    original_st_){original_st = original_st_;}
	virtual int    getOriginal_st() const {return original_st;}

	virtual void   setModelled_wd(double modelled_wd_){modelled_wd = modelled_wd_;}
	virtual double getModelled_wd() const {return modelled_wd;}
	virtual void   setModelled_ws(double modelled_ws_){modelled_ws = modelled_ws_;}
	virtual double getModelled_ws() const {return modelled_ws;}
	virtual void   setModelled_st(int    modelled_st_){modelled_st = modelled_st_;}
	virtual int    getModelled_st() const {return modelled_st;}

	virtual void   setSz0(double sz0_){sz0 = sz0_;}
	virtual double getSz0() const {return sz0;}
	virtual void   setSy0(double sy0_){sy0 = sy0_;}
	virtual double getSy0() const {return sy0;}

	virtual double getWindDirectionALaMathematician(double blowing_direction_measured_from_north_with_clockwise_in_degrees) const;

	virtual void   setOpenAngle(double in_degrees){open_angle = in_degrees;}
	virtual double getOpenAngle() const {return open_angle;}

	virtual double getTrigHalfOpenAngle(Trig tr);

	virtual bool   isBadData() const;

	enum UrbanOrRural {URBAN, RURAL};
	virtual double power4ws (int st_, UrbanOrRural uor) const;

	// Runtime options
	virtual void   setConcentrationUnitFlag(bool concentration_unit_flag_){concentration_unit_flag = concentration_unit_flag_;}
	virtual bool   getConcentrationUnitFlag() const {return concentration_unit_flag;}
	virtual void   setPrefix(string casename_prefix_) {casename_prefix = casename_prefix_;}
	virtual string getPrefix() const {return casename_prefix;}
	virtual void   setProgressPeekStep(int progress_peek_step_) {progress_peek_step = progress_peek_step_;}
	virtual int    getProgressPeekStep() const {return progress_peek_step;}

	virtual void   setMinimumIterations(int minimum_iterations_) {minimum_iterations = minimum_iterations_;}
	virtual int    getMinimumIterations() const {return minimum_iterations;}
	virtual void   setMaximumIterations(int maximum_iterations_) {maximum_iterations = maximum_iterations_;}
	virtual int    getMaximumIterations() const {return maximum_iterations;}
	virtual void   setToleranceRatio(double tolerance_ratio_) {tolerance_ratio = tolerance_ratio_;}
	virtual double getToleranceRatio()    const {return tolerance_ratio;}

	//
	// List level option communicators
	//
	// Level	|	Description
	// -------------------------
	//     1	|	Essential
	//     2	|	Moderate
	//     3	|	Full
	//

	enum    ListLevel {ESSENTIAL, MODERATE, FULL};
	virtual void       setListLevel(int list_level_);
	virtual ListLevel  getListLevel() const {return list_level;}

	Simulator();
private:
	ListLevel list_level;
public:
	friend class NumIntegrator; friend class Trapezoid; friend class Simpson; friend class ErrorSimpson; friend class SimulationFactory;

	virtual void scan ();
	virtual void print();
	virtual void lists();
	virtual void exec (bool new_met, bool new_src){}
	virtual void   setResult(Rcp::Result res, double val);
	virtual double getResult(Rcp::Result res) const;

	virtual Met* getMetPtr(){return ptr_met;}
	virtual Rcp* getRcpPtr(){return ptr_rcp;}
	virtual Src* getSrcPtr(){return ptr_src;}
	virtual Sgm* getSgmPtr(){return ptr_sgm;}
	virtual Pat* getPatPtr(){return ptr_pat;}
	virtual Ran* getRanPtr(){return ptr_ran;}
	virtual Blg* getBlgPtr(){return ptr_blg;}

	virtual void setTimeSeriesFlag(bool is_time_series_){is_time_series = is_time_series_;}
	virtual bool isTimeSeries() const {return is_time_series;}
	virtual int  getMaxDailyCodes()  const {return tim-> getMaxDailyCodes();}
	virtual int  getMaxHourlyCodes() const {return tim->getMaxHourlyCodes();}
	virtual void assembleResult(int counter);

	virtual void addSrcContrib(double corr_factor = 1.0);

	virtual ~Simulator();
};

#endif
